// CharDeque.h

#pragma once

#ifndef _CHAR_DEQUE_H_
#define _CHAR_DEQUE_H_

#include "FastDeque.h"


//--------------------------------------------------------------
// CharDequeImpl
//
// warning: this deque class copy elements using memcpy
//--------------------------------------------------------------

template < typename T, typename DequeAllocator, unsigned chunk_size >
class PodDequeImpl
{
public:
	class forward_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef T * pointer;
		typedef T & reference;
		//};

		friend FastDequeImpl;
	private:
		forward_iterator ( T * current, T * partial_end )
			: current_ ( current )
			, partial_end_ ( partial_end )
		{}

	public:
		forward_iterator()
			: current_ ( 0 )
			, partial_end_ ( 0 )
		{}

		T & operator*() const
		{
			return * current_;
		}

		const T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const forward_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const forward_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		forward_iterator & operator++ () // preincrement
		{
			++current_;
			if ( current_ == partial_end_ )
			{
				T ** p_part_next = reinterpret_cast<T**>(partial_end_);
				T * part_next = * p_part_next;
				if ( part_next )
				{
					current_ = part_next + begin_data_index_;
					partial_end_ = part_next + end_data_index_;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return *this;
		}

		forward_iterator operator++ ( int ) // postincrement
		{
			forward_iterator retValue = *this;
			++current_;
			if ( current_ == partial_end_ )
			{
				T ** p_part_next = reinterpret_cast<T**>(partial_end_);
				T * part_next = * p_part_next;
				if ( part_next )
				{
					current_ = part_next + begin_data_index_;
					partial_end_ = part_next + end_data_index_;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return retValue;
		}

		//forward_iterator & operator-- () // predecrement
		//{
		//	--current_;
		//	return *current_;
		//}

		//forward_iterator operator-- ( int ) // postdecrement
		//{
		//	forward_iterator retValue = *this;
		//	--*this;
		//	return retValue;
		//}

	private:
		T * current_;
		T * partial_end_;
	};

	class const_forward_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef const T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef const T * pointer;
		typedef const T & reference;
		//};

	private:
		const_forward_iterator ( T * current, T * partial_end )
			: current_ ( current )
			, partial_end_ ( partial_end )
		{}

	public:
		const_forward_iterator()
			: current_ ( 0 )
			, partial_end_ ( 0 )
		{}

		const T & operator*() const
		{
			return * current_;
		}

		const T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const const_forward_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_forward_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_forward_iterator & operator++ () // preincrement
		{
			++current_;
			if ( current_ == partial_end_ )
			{
				T ** p_part_next = reinterpret_cast<T**>(partial_end_);
				T * part_next = * p_part_next;
				if ( part_next )
				{
					current_ = part_next + begin_data_index_;
					partial_end_ = part_next + end_data_index_;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return *this;
		}

		const_forward_iterator operator++ ( int ) // postincrement
		{
			forward_iterator retValue = *this;
			++current_;
			if ( current_ == partial_end_ )
			{
				T ** p_part_next = static_cast<T**>(partial_end_);
				T * part_next = * p_part_next;
				if ( partial_next )
				{
					current_ = part_next + begin_data_index_;
					partial_end_ = part_next + end_data_index_;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return retValue;
		}

		//forward_iterator & operator-- () // predecrement
		//{
		//	--current_;
		//	return *current_;
		//}

		//forward_iterator operator-- ( int ) // postdecrement
		//{
		//	forward_iterator retValue = *this;
		//	--*this;
		//	return retValue;
		//}

	private:
		T * current_;
		T * partial_end_;
	};

	typedef forward_iterator iterator;
	typedef const_forward_iterator const_iterator;

// part

	struct part
	{
		part ( T * data_begin, T * data_end )
			: data_begin_ ( data_begin )
			, data_end_ ( data_end )
		{}

		T * begin()
		{
			return data_begin_;
		}

		T * end()
		{
			return data_end_;
		}

	private:
		T * data_begin_;
		T * data_end_;
	};

	class part_iterator
	{
	public:
		part_iterator ( T * current_part, T * last_part, T * last_part_data_end )
			: current_part_ ( current_part )
			, last_part_ ( last_part )
			, last_part_data_end_ ( last_part_data_end )
		{}

		bool operator == ( const part_iterator & rhs ) const
		{
			return current_part_ == rhs.current_part_;	
		}

		bool operator != ( const part_iterator & rhs ) const
		{
			return current_part_ != rhs.current_part_;	
		}

		part operator*()
		{
			if ( current_part_ != last_part_ )
				return part ( current_part_ + begin_data_index_, current_part_ + end_data_index_ );
			else
				return part ( current_part_ + begin_data_index_, last_part_data_end_ );
		}

		part_iterator & operator++ () // preincrement
		{
			T ** p_part_next = reinterpret_cast<T**>(current_part_ + end_data_index_);
			current_part_ = *p_part_next;
			return *this;
		}

	private:
		T * current_part_;
		T * last_part_;
		T * last_part_data_end_;
	};

// constructor

	PodDequeImpl()
		: first_ ( 0 )
		, end_ ( 0 )
		, last_ ( 0 )
	{}

	void clear()
	{
		T * next = first_;
		while ( next )
		{
			T ** p_part_next = reinterpret_cast<T**>(next + end_data_index_);
			T * part_next = *p_part_next;
			deallocate ( (byte*)next );
			next = part_next;
		}
		first_ = 0;
		end_ = 0;
		last_ = 0;
	}

	~PodDequeImpl()
	{
		clear();
	}

private:
	static byte * allocate()
	{
		return DequeAllocator::allocate();
		//byte * address = (byte*)FixedSmallObjAllocator<sizeof(T)*chunk_size_,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		////byte * address = new byte[sizeof(T)*chunk_size_];
		//return (byte*)address;
	}

	static void deallocate ( byte * address )
	{
		//FixedSmallObjAllocator<sizeof(T)*chunk_size_,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( address );
		////delete address;
		return DequeAllocator::deallocate ( address );
	}

public:
	forward_iterator forward_begin()
	{
		if ( first_ )
			return forward_iterator ( first_ + begin_data_index_, first_ + end_data_index_ );
		else
			return forward_iterator ( 0, 0 );
	}

	const_forward_iterator forward_begin() const
	{
		if ( first_ )
			return const_forward_iterator ( first_ + begin_data_index_, first_ + end_data_index_ );
		else
			return const_forward_iterator ( 0, 0 );
	}

	forward_iterator forward_end()
	{
		if ( last_ )
			return forward_iterator ( last_, end_ - ( chunk_size_ - end_data_index_ ) );
		else
			return forward_iterator ( 0, 0 );
	}

	const_forward_iterator forward_end() const
	{
		if ( last_ )
			return const_forward_iterator ( last_, end_ - ( chunk_size_ - end_data_index_ ) );
		else
			return const_forward_iterator ( 0, 0 );
	}

	forward_iterator begin() { return forward_begin(); }
	const_forward_iterator begin() const { return forward_begin(); }
	forward_iterator end() { return forward_end(); }
	const_forward_iterator end() const { return forward_end(); }

	const T & front() const
	{
		return *(first_ + begin_data_index_);
	}

	T & front()
	{
		return *(first_ + begin_data_index_);
	}

	const T & back() const
	{
		return *(last_ - 1 );
	}

	T & back()
	{
		return *(last_ - 1 );
	}

	part_iterator part_begin() const
	{
		return part_iterator ( first_, end_ - chunk_size_, last_ );
	}

	part_iterator part_end() const
	{
		return part_iterator ( 0, 0, 0 );
	}

private:
	void alloc_next_part()
	{
		T * new_part = reinterpret_cast<T*>(allocate());
		if ( end_ )
		{
			T ** prev_next_part = reinterpret_cast<T**> ( end_ - ( chunk_size_ - end_data_index_ ) );
			*prev_next_part = new_part;
		}
		T ** next_prev_part = reinterpret_cast<T**>(new_part);
		*next_prev_part = end_ ? end_ - chunk_size_ : 0;
		T ** next_next_part = reinterpret_cast<T**>(new_part + end_data_index_);
		*next_next_part = 0;
		if ( !first_ )
			first_ = new_part;
		last_ = new_part + begin_data_index_;
		end_ = new_part + chunk_size_;
	}

public:
	void push_back ( T value )
	{
		//INVARIANT();

		//#ifdef _DEBUG
		//		unsigned debug_size_before = eval_size();
		//#endif // _DEBUG

		if ( last_ == 0
			|| last_ == end_ - ( chunk_size_ - end_data_index_ ) )
		{
			alloc_next_part();
		}
		*last_++ = value;

		//#ifdef _DEBUG
		//		unsigned debug_size_after = eval_size();
		//		XASSERT ( debug_size_before + 1 == debug_size_after );
		//#endif // _DEBUG


		//INVARIANT();
	}

	void push_back ( const T * begin, const T * end )
	{
		//INVARIANT();

		size_t size = end - begin;
		if ( size && !first_ )
			alloc_next_part();
loop:
		size_t available = ( end_ - ( chunk_size_ - end_data_index_ ) ) - last_;
		if ( size <= available )
		{
			memcpy ( last_, begin, sizeof(T)*size );
			last_ += size;
			//INVARIANT();
			return;
		}
		memcpy ( last_, begin, available );
		begin += available;
		size -= available;
		alloc_next_part();
		goto loop;
	}

	//template<typename T2>
	//void push_back ( T2 it, T2 end )
	//{
	//	for ( ; it != end; ++it )
	//	{
	//		push_back ( *it );
	//	}
	//}

	bool empty() const
	{
		return first_ == last_;
	}

	//size_t size() const
	//{
	//	return last_ - first_;
	//}

	//T& operator[] ( unsigned index )
	//{
	//	return first_[index];
	//}

	unsigned eval_size() const
	{
		unsigned size = 0;
		if ( first_ )
		{
			T * part = first_;
			//XASSERT ( next == prev_part );
			while ( part )
			{
				T * next_part = *reinterpret_cast<T**>(&part[end_data_index_]);
				if ( next_part )
					size += end_data_index_ - begin_data_index_;
				else
					size += last_ - ( part + begin_data_index_ );
				part = next_part;
			}
		}
		return size;
	}


private:
	static const unsigned chunk_size_ = chunk_size;
	static const unsigned begin_data_index_ = DIV_ROUND_UP(sizeof(T*),sizeof(T));
	static const unsigned end_data_index_ = chunk_size_ - DIV_ROUND_UP(sizeof(T*),sizeof(T));
	T * first_; // pointer to beginning of array
	T * end_; // pointer to current end of sequence
	T * last_; // pointer to current end of sequence
};

//------------------------------------------------------------------------------
// CharDeque
//------------------------------------------------------------------------------

typedef PodDequeImpl<char,MediumDequeAllocator,MediumDequeAllocator::page_size/sizeof(char)> CharDeque;

//------------------------------------------------------------------------------

#endif // _CHAR_DEQUE_H_